import os
import logging
import json
import sys
import re
import webbrowser
import hashlib
import difflib

from .parser import ParseNotebook

try:
    import nbconvert
except:
    nbconvert = None

import nbformat

from .styles import process_css

# TODO
# [ ] Could add a kill buffer hook to shutdown server when all nei buffers closed.
# [ ] Add Yank hook.

def serialize_binary_message(msg):
    """serialize a message as a binary blob
    Header:
    4 bytes: number of msg parts (nbufs) as 32b int
    4 * nbufs bytes: offset for each buffer as integer as 32b int
    Offsets are from the start of the buffer, including the header.
    Returns
    -------
    The message serialized to bytes.
    """
    from jupyter_client.jsonutil import date_default
    import struct
    # don't modify msg or buffer list in-place
    msg = msg.copy()
    buffers = list(msg.pop('buffers'))
    if sys.version_info < (3, 4):
        buffers = [x.tobytes() for x in buffers]
    bmsg = json.dumps(msg, default=date_default).encode('utf8')
    buffers.insert(0, bmsg)
    nbufs = len(buffers)
    offsets = [4 * (nbufs + 1)]
    for buf in buffers[:-1]:
        offsets.append(offsets[-1] + len(buf))
    offsets_buf = struct.pack('!' + 'I' * (nbufs + 1), nbufs, *offsets)
    buffers.insert(0, offsets_buf)
    return b''.join(buffers)



class Cell(object):
    "Object representing a code or markdown cell"

    code_pretty_formatter = "# In[{prompt}]\n{source}\n\n"
    md_pretty_formatter = '\"\"\"\n{source}\n\"\"\"\n\n'

    code_prompt_formatter = "# In[{prompt}]\n{source}"
    md_prompt_formatter = '\"\"\"\n{source}\n\"\"\"'

    # For quick startswith line matching
    code_startswith = "# In["
    md_startswith = '\"\"\"'

    # To extract the prompt number (if any)
    code_prompt_regexp = re.compile("^# In\[(\d*| )\]")

    def __init__(self, mode, source, input=None, prompt=None, outputs=[]):
        self.mode = mode
        self.source = source
        self.prompt = prompt
        self.input = input

        self.outputs = outputs # List of nbformat output nodes

    @property
    def data(self):
        return {'mode':self.mode,
                'source':self.source,
                'prompt':self.prompt}

    @property
    def text(self):
        if self.mode == 'code':
            prompt = ' ' if self.prompt is None else self.prompt
            return self.code_prompt_formatter.format(source=self.source, prompt=prompt)
        elif self.mode == 'markdown':
            return self.md_prompt_formatter.format(source=self.source)

    @property
    def hash(self):
        return hash((self.prompt, self.mode,self.source.strip()))

    def unquote_markdown(self, source):
        """
        Replace quoted triple quotes i.e zero width space separated
        double quotes ("​"​") with unquoted versions
        """
        return source.replace('"​"​"', '"""')

    def node(self, cleared=False):
        "Return cell as a notebook node"
        out_nodes = []
        if not cleared:
            for output in self.outputs: # TODO: Support other outputs
                out_nodes.append(output)

        if self.mode == 'code':
            return nbformat.v4.new_code_cell(self.source, outputs=out_nodes)
        elif self.mode == 'markdown':
            return nbformat.v4.new_markdown_cell(self.unquote_markdown(self.source))

    def clear_output(self):
        self.outputs = []

    def __repr__(self):
        return "%scell(%s)" % (self.mode, self.source[:50])


class Cells(object):
    "Notebook state is represented as a list of Cell objects"

    def __init__(self, text=None, cells=[]):
        if text is None:
            self.cells = cells
        else:
            self.load(ParseNotebook.extract_cells(text, Cell))

    @property
    def hashes(self):
        return [cell.hash for cell in self.cells]

    @property
    def text(self):
        "Return the textual representation of the cells"
        return ''.join([cell.text for cell in self.cells])

    @property
    def pretty(self):
        print("PRETTY NOT IMPLEMENTED")
        return self.text

    def load(self, dicts):
        "Load cells from cell dictionaries"
        if dicts is None: return
        self.cells = [Cell(**cdict) for cdict in dicts]

    def by_line(self, line_number, offset=False):
        "Given a line number find the corresponding cell number with offset if requested"
        md_closed, cell_count, boundary = True, -1, 0
        for lineno, line in enumerate(str(self.mirrorbuffer).splitlines(keepends=True)):
            if line.startswith(Cell.code_startswith):
                cell_count += 1
                boundary = lineno
            if line.startswith(Cell.md_startswith) and md_closed:
                cell_count += 1
                boundary = lineno
                md_closed = not md_closed
            elif line.startswith(Cell.md_startswith):
                md_closed = not md_closed
            if line_number == lineno+1:
                cell_no = cell_count if cell_count >= 0 else None
                delta = lineno - boundary
                return (cell_no, delta) if offset else cell_no

        logging.info('WARNING: Line number out of bounds')

    def find_cell(self, prompt):
        "Given a prompt number, find the corresponding cell, if any"
        for cell in self.cells:
            if cell.prompt == prompt:
                return cell
        logging.info("WARNING: Could not find cell with prompt %s" % prompt)

    def cell_position(self, cell):
        "Given a cell, return its position"
        for pos, candidate in enumerate(self.cells):
            if candidate == cell:
                return pos

    # Commands (that don't need to talk to the client)

    def write_notebook(self, connection, mode, filename):
        assert mode in ['python', 'cleared', 'full-notebook', 'html']
        if mode == 'python':
            data = self.pretty
        else:
            cleared = (mode == 'cleared')
            nb = nbformat.v4.new_notebook()
            nb['cells'] = [cell.node(cleared=cleared) for cell in self.cells]
            if self.metadata is not None:
                nb['metadata'] = self.metadata

        if mode == 'html':
            if nbconvert is None:
                logging.info("nbconvert not available for HTML conversion")
                return
            (data, resources) = nbconvert.exporters.export(nbconvert.HTMLExporter, nb)
        with open(filename, 'w') as f:
            if mode in ['cleared', 'full-notebook']:
                nbformat.write(nb, f)
            else:
                f.write(data)
        return {'cmd':'write_complete', 'data':self.name}

    def _open_browser(self, url):
        chosen_browser = self.config['browser']
        fail_msg = {'cmd':'user_message',
                    'data':'Could not open requested browser %r' % chosen_browser}
        try:
            browser_obj = webbrowser.get(chosen_browser)
        except:
            return fail_msg

        success = browser_obj.open(url)
        if not success: return fail_msg

    def view_notebook(self, connection):
        filename = os.path.join(self.STATIC_PATH, 'view.html')
        self.write_notebook(connection, 'html', filename)
        return self._open_browser("http://localhost:8000/view.html")

    def view_browser(self, connection, ws_port=9999):

        if ws_port == 9999:
            url = "http://localhost:8000/index.html"
        else:
            url = "http://localhost:8000/index.html?ws-port={port}".format(port=ws_port)

        if connection is None:
            return self._open_browser(url)


class OutputMessage(object):
    """
    Class to process and filter output messages sent to the frontend.
    """

    @classmethod
    def process(cls, node):
        out_msgs = cls._output_messages(node)
        return cls._filter_outputs(out_msgs)

    @classmethod
    def _filter_outputs(cls, outputs):
        "Remove plaintext mimetype if html mimetype available"
        if outputs == []: return []
        mimes, data = zip(*outputs)
        exclude = all(m in mimes for m in ['text/plain', 'text/html'])
        return [(mime, val) for (mime, val) in zip(mimes, data)
                if mime != ('text/plain'if exclude else None)]


    @classmethod
    def _output_messages(cls, node):
        "Given a notebook node, generate output messages"
        mime_types = ['text/plain', 'text/html', 'text/ansi',
                      'application/javascript']
        output_type = node['output_type']
        # Assuming notebook node is iterated in order.
        if output_type == 'stream':
            text = node['text']
            if node['name'] == 'stderr':
                text = "<p style='color:red'>{text}</p>".format(text=text)
            return [('text/html', text)]
        elif output_type == 'error': # Handle warnings
            return[('text/ansi', node['traceback'])]

        outputs = []
        for mtype, data in node['data'].items():
            if mtype not in mime_types: continue
            outputs.append((mtype, data))
        return outputs



class Notebook(Cells):
    """
    Extends cells with methods that operate on them and communicate with the client

    PROTOCOL

    In browser, 'cmd' can be:
    'add_cell','update_cell', 'remove_cells', 'clear_cell_output'

    In editor, 'cmd' can *also* be:
    'exec_cell_by_line','exec_cell_by_line'
    which then updates the display
    """

    STATIC_PATH = None

    def __init__(self, name=None, **kwargs):
        super(Notebook, self).__init__(**kwargs)
        self.name = name
        self.metadata = None
        self._last_dispatch = None # Used for debugging and testing
        self.mirrorbuffer = MirrorBuffer()
        self.commands = {'view_browser': self.view_browser, # Opens browser connection
                         # Commands that do no need a browser connection
                         'update_config':    self.update_config,
                         'scroll_by' :       self.scroll_by,
                         'scroll_to_line' :  self.scroll_to_line,

                         # Used for debugging
                         'terminate' :       self.terminate,
                         'server_info' :     self.server_info,

                         'write_notebook':   self.write_notebook,
                         'start_mirror':     self.start_mirror,
                         'mirror':           self.mirror,
                         'mirroring_error':  self.mirroring_error,
                         'hold_mode':        self.hold_mode,
                         # Cell commands
                         'add_cell':            self.add_cell,
                         'update_cell_outputs': self.update_cell_outputs,
                         'update_cell_input':   self.update_cell_input,
                         'clear_cell_output':   self.clear_cell_output,
                         # Multi-cell commands
                         'load_from_file':         self.load_from_file,
                         'insert_file_at_point':   self.insert_file_at_point,
                         'reorder_cells':          self.reorder_cells,
                         'clear_notebook':         self.clear_notebook,
                         'clear_all_cell_outputs': self.clear_all_cell_outputs,
                         # IO commands
                         'download_python_notebook':  self.download_python_notebook,
                         'download_cleared_notebook': self.download_cleared_notebook,
                         'download_full_notebook':    self.download_full_notebook,
                         'view_notebook':             self.view_notebook,
                         'update_theme':              self.update_theme,
                         'display_code':              self.display_code,
                         'display_all_code':          self.display_all_code,
                         # Derived 'by_line' methods
                         'clear_cell_output_by_line': self.clear_cell_output_by_line,
                         }
        self.css = "" # Last CSS sent to browser
        self.config = {'browser':'firefox'}
        self.code_visible = True


    def message(self, connection, command, args, buffers=[]):
        if connection is not None:
            if buffers == []:
                connection.write_message({'cmd':command, 'args':args, 'name':self.name})
            else:
                binary = serialize_binary_message(
                    {'cmd':'comm_msg', 'args':args, 'name':self.name, 'buffers': buffers})
                connection.write_message(binary, binary=True)
        else:
            logging.info("WARNING: Command %s sent to non-connection" % command)

    def dispatch(self, connection, payload):
        """
        Dispatches to commands. If commands return a JSON serializable
        value, it is passed to the editor if possible (most commands
        return None)
        """
        cmd, args = payload['cmd'], payload['args']
        if cmd != 'server_info':
            self._last_dispatch = payload # Used for debugging/testing
        return self.commands[cmd](connection, **(args if args is not None else {}))

    def update_config(self, connection, config):
        self.config = config

    def terminate(self, connection): # Used for debugging
        logging.info("ERROR: Termination requested")

    def server_info(self, connection):
        # Used for debugging and testing by outputting info on browser connection
        self.message(connection, 'server_info',
                     {'text': self.text,
                      'last_dispatch': self._last_dispatch})

    def scroll_by(self, connection, offset):
        self.message(connection, 'scroll_by', {'offset': offset})

    def scroll_to_line(self, connection, line):
        info = self.by_line(line, offset=True)
        if info is None: return
        position, offset = info
        self.message(connection, 'scroll_to', {'position':position, 'line':offset})

    def clear_cell_output(self, connection, position):
        self.cells[position].clear_output()
        self.message(connection, 'clear_cell_output', {'position':position})

    def clear_cell_output_by_line(self, connection, line_number):
        position = self.by_line(line_number)
        if position is None: return
        self.clear_cell_output(connection, position)


    def remove_cell(self, connection, position):
        try:
            del self.cells[position]
        except:
            logging.info("WARNING: Could not remove cell %d when cells list is %d long"
                         % (position, len(self.cells)))
        self.message(connection, 'remove_cell', {'position':position})

    def add_cell(self, connection, source, input=None,
                 mode='code', prompt=None, position=None, outputs=[]):
        cell = Cell(mode=mode, source=source, input=input, prompt=prompt, outputs=outputs)
        if position is None:
            self.cells.append(cell)
        else:
            self.cells.insert(position, cell)

        args = {'mode':mode, 'source':source, 'input':input,
                'position':position, 'outputs':[]}
        if outputs is not None:
            filtered_outputs = []
            for cell_output in outputs:
                filtered_outputs.extend(OutputMessage.process(cell_output))
            args['outputs'] = filtered_outputs

        self.message(connection, 'add_cell', args)


    def update_cell_outputs(self, connection, position, outputs):
        "Call with outputs of None to simply update the prompt"
        cell = self.cells[position]
        if outputs is not None:
            cell.outputs.append(outputs)
        if not isinstance(outputs, (list, type(None))): # I.e an output node
            # Filtering output messages
            outputs =  OutputMessage.process(outputs)

        self.message(connection, 'update_cell_outputs',{'position':position,
                                                        'outputs': outputs,
                                                        'prompt': cell.prompt})



    def update_cell_input(self, connection, source, position, input=None):
        self.cells[position].source = source
        self.cells[position].input = input
        self.message(connection, 'update_cell_input',{'position':position,
                                                      'source':source,
                                                      'input': input})

    #=================================================#
    # Commands that operate on multiple cells at once #
    #=================================================#

    def reload(self, connection):
        """
        Called when the browser tab is initially opened or reloaded and
        the notebook has cell state.
        """
        self.message(connection, 'clear_notebook', {})
        for pos, cell in enumerate(self.cells):
            # TODO: Pass prompt to JS
            self.message(connection, 'add_cell', {'mode':cell.mode,
                                                  'source':cell.source,
                                                  'input':cell.input})
            if cell.mode == 'code' and cell.outputs:
                for output in cell.outputs:
                    filtered = OutputMessage.process(output)
                    self.message(connection,'update_cell_outputs', {'position':pos,
                                                                    'outputs':filtered})


    def mirror(self, connection, start, end, length, added, size, md5=None):
        mirroring_error = self.mirrorbuffer(start, end, length, added, size, md5)
        if not self.mirrorbuffer.hold:
            cells = ParseNotebook.extract_cells(str(self.mirrorbuffer), Cell)
            src = Notebook(cells=list())
            src.load(cells)
            SyncNotebooks.sync(connection, src, self)
            if mirroring_error:
                return {'cmd':'mirroring_error', 'data':mirroring_error}

    def mirroring_error(self, connection, editor_text, mirror_text):
        d = difflib.Differ()
        diff = d.compare(editor_text.splitlines(), mirror_text.splitlines())
        raise AssertionError('Mirroring error with diff:\n\n%s' % '\n'.join(diff))

    def start_mirror(self, connection, text):
        self.mirrorbuffer.clear()
        self.mirror(connection, 0, len(text), len(text), text, len(text))

    def hold_mode(self, connection, mode):
        assert mode in ['on','off','auto']
        self.mirrorbuffer.hold_mode = mode

    def insert_file_at_point(self, connection, filename, text, line_number):
        insert_cell_pos = self.by_line(line_number)
        # Mirror buffer expected to be already updated by mirror insertion from editor
        buffer_text = str(self.mirrorbuffer)
        cells = ParseNotebook.extract_cells(buffer_text, Cell)
        src = Notebook(cells=list())
        src.load(cells)

        with open(filename, 'r') as f:
            nb = nbformat.v4.reads(f.read())

        positions = range(insert_cell_pos, insert_cell_pos + len(nb.cells))
        assert len(positions) == len(nb.cells)
        for insert_pos, nb_cell in zip(positions, nb.cells):
            if nb_cell.cell_type == 'code':
                src.cells[insert_pos].outputs = nb_cell.outputs

        SyncNotebooks.sync(connection, src, self)

    def load_from_file(self, connection, json_string, filename, buffer_text=None):
        "Clears any existing cells and replaces them with ones loaded from file"
        self.clear_notebook(connection)

        with open(filename, 'r') as f:
            nb = nbformat.v4.reads(f.read())

        self.metadata = nb.metadata
        dict_cells = json.loads(json_string)

        if buffer_text:
            valid = ParseNotebook.validate_notebook(nb, dict_cells, buffer_text, Cell)
            if not valid:
                return {'cmd':'load_validated',
                        'data':{'filename':filename, 'valid':False}}

        for (pos, (dict_cell, nb_cell)) in enumerate(zip(dict_cells, nb.cells)):
            if nb_cell.cell_type == 'code':
                dict_cell['outputs'] = nb_cell.outputs

            dict_cell['position'] = pos
            self.add_cell(connection, **dict_cell)

        return {'cmd':'load_validated',
                'data':{'filename':filename, 'valid':True}}

    def reorder_cells(self, connection, positions):
        assert len(positions) == len(self.cells)
        reordered = []
        for _, target in sorted([(pos,ind) for (ind,pos) in enumerate(positions)]):
            reordered.append(self.cells[target])
        self.cells = reordered
        self.message(connection,'reorder_cells', {'positions':positions})


    def clear_all_cell_outputs(self, connection):
        for pos in range(len(self.cells)):
            self.clear_cell_output(connection, pos)


    def clear_notebook(self, connection):
        self.cells = []
        self.message(connection,'clear_notebook', {})


    #=============#
    # IO Commands #
    #=============#

    # TODO: Can simplify/unify the download commands
    def download_python_notebook(self, connection):
        self.message(connection, 'download_python_notebook',
                     {'filename':'python_notebook.py',
                      'data':self.pretty})

    def download_cleared_notebook(self, connection):
        nb = nbformat.v4.new_notebook()
        nb['cells'] = [cell.node(cleared=True) for cell in self.cells]
        self.message(connection, 'download_python_notebook',
                     {'filename':'cleared_notebook.ipynb',
                      'data':nbformat.writes(nb)})

    def download_full_notebook(self, connection):
        nb = nbformat.v4.new_notebook()
        nb['cells'] = [cell.node(cleared=False) for cell in self.cells]
        self.message(connection, 'download_python_notebook',
                     {'filename':'full_notebook.ipynb',
                      'data':nbformat.writes(nb)})


    def update_theme(self, connection, css):
        if css is None: # Use stored CSS if the tab is reloaded
            css = self.css
        else:
            self.css = css
        lines = css.splitlines()

        self.message(connection, 'update_theme', {'css':"\n".join(lines[1:-1])})

    def display_code(self, connection, visible, line_number):
        # Argument visible can be True, False or "toggle"
        pos = self.by_line(line_number)
        if self.cells[pos].mode == 'code':
            self.message(connection, 'display_code',
                         {'pos': pos, 'visible': visible})

    def display_all_code(self, connection, visible):
        if visible == "toggle":
            self.code_visible = not self.code_visible
            visible = self.code_visible

        positions = []
        for cell in self.cells:
            if cell.mode == 'code':
                positions.append(self.cell_position(cell))

        for pos in positions:
            self.message(connection, 'display_code',
                         {'pos': pos, 'visible': visible})



class ExecutableNotebook(Notebook):
    """
    ExecutableNotebook is a notebook with executable state i.e a Python
    kernel along with methods for executing code.
    """
    def __init__(self, executor_init, name=None, **kwargs):
        self.executor_init = executor_init
        self.executor = None # Initialized by start_kernel
        self.completion_info = None
        super(ExecutableNotebook, self).__init__(name=name, **kwargs)
        self.exec_commands = {
            'exec_silently':    self.exec_silently,
            'exec_cell':        self.exec_cell,
            'exec_cell_by_line':self.exec_cell_by_line,
            # For JS testing only...
            'add_cell_exec':    self.add_cell_exec,
            'comm_open':        self.comm_open,    # Request from JS to open comm
            'comm_msg' :        self.comm_msg, # Message send from JS
            'start_kernel':     self.start_kernel,
            'interrupt_kernel': self.interrupt_kernel,
            'restart_kernel':   self.restart_kernel,
            'complete':         self.complete
        }
        self.commands.update(self.exec_commands)

    def complete(self, connection, line_number, line_context, position):
        cell_index =  self.by_line(line_number)
        cell_source = self.cells[cell_index].source

        relative_position = 0
        for line in cell_source.split("\n"):
            if line.startswith(line_context):
                relative_position += len(line_context)
                break
            relative_position += len(line) +1 # +1 to compensate for newlines

        self.completion_info = {'cell_source':cell_source,
                                'position': position,
                                'relative_position':relative_position}
        self.executor.complete(cell_source, relative_position)

    def dispatch(self, connection, payload):
        kernel_required = [cmd for cmd in self.exec_commands if cmd != 'start_kernel']
        if self.executor is None and (payload['cmd'] in kernel_required):
            logging.info("WARNING: Kernel needs to be started before execution can occur")
            return
        return super(ExecutableNotebook, self).dispatch(connection, payload)

    def exec_silently(self, connection, code):
        self.executor(code, stop_on_error=False, cell=None, silent=True)

    def comm_open(self, connection, target_name, comm_id, data=None, metadata=None):
        # Request to open a comm from JS
        self.executor.comms.comm_open(comm_id=comm_id,
                                      target_name=target_name,
                                      data=data, metadata=metadata)

    def comm_msg(self, connection, target_name, comm_id, data, metadata=None):
        # Message from JS comm to Python
        self.executor.comms.comm_msg(comm_id=comm_id,
                                     target_name=target_name,
                                     data=data, metadata=metadata)

    def exec_cell(self, connection, position):
        if position >= len(self.cells):
            logging.info("WARNING: Position %d is out of bounds" % position)
            return
        cell = self.cells[position]
        if cell.mode == 'markdown': return
        if cell.outputs:
            self.clear_cell_output(connection, position)

        self.executor(cell.source, stop_on_error=True, cell=cell)


    def add_cell_exec(self, connection, source, input=None,
                      mode='code', prompt=None, position=None):
        """
        Needed to solve concurrency issues in JS where add_cell messages arrive
        before the corresponding executions. Should only be used for JS tests.
        """
        position = len(self.cells) if position is None else position
        self.add_cell(connection, source, input, mode, prompt, position)
        self.exec_cell(connection, position)


    def exec_cell_by_line(self, connection, line_number):
        position =  self.by_line(line_number)
        if position is None: return
        self.exec_cell(connection, position)

    def start_kernel(self, connection, cwd=None, executable=None):
        exec_cls, name, queue = self.executor_init
        self.executor  = exec_cls(name, queue)
        self.executor.start(cwd, executable)

    def interrupt_kernel(self, connection):
        self.executor.interrupt_kernel()

    def restart_kernel(self, connection):
        self.executor.restart_kernel()


    def update_theme(self, connection, css):
        if self.executor and css:
            lines = css.splitlines()
            process_css(self, connection, "\n".join(lines[1:-1]))
        super(ExecutableNotebook, self).update_theme(connection, css)




class MirrorBuffer(object):

    raise_on_mismatch = True # Useful for debugging

    def __init__(self, hold_mode='off'):
        self.buff = ''
        self._hold = False
        self.hold_mode = hold_mode # May be 'on', 'off' or 'auto'

    def clear(self):
        self.buff = ''

    @property
    def hold(self):
        """
        Sync can't distinguish a cut and paste (i.e a move) from an
        initial deletion without the concept of a 'hold'. The hold sapplies
        if a cell prompt initially appears to be deleted
        """
        return self._hold


    def _check_hold(self, start, length, added):
        if self.hold_mode in ['on','off']:
            self._hold = (self.hold_mode == 'on')
        elif added != "" or length < 5:
            self._hold = False
        else:
            deletion = self.buff[start-1:start+length-1]
            self._hold = deletion.strip().startswith(Cell.code_startswith)

    def __call__(self, start, end, length, added, size, md5=None):
        self._check_hold(start, length, added)
        if length-1 >= 0:
            self.buff = self.buff[:start-1] + self.buff[start+length-1:]
        self.buff = self.buff[:start-1] + added + self.buff[start-1:]

        # Use to simulate random corruption of mirror buffer
        # import random
        # if random.random() < 0.01:
        #     index = random.randint(0, len(self.buff)-1)
        #     self.buff = self.buff[:index] + 'X' + self.buff[index:]

        if md5 is not None:
            buffer_hash = hashlib.md5(self.buff.encode('utf-8')).hexdigest()
            if (md5 != buffer_hash):
                msg = "MD5 mismatch between buffer (len %d) and mirrored contents (len %d)."
                formatted = msg % (size, len(self.buff))
                logging.warning(msg)
                # Return the buffer to send a message back to the editor for debug
                return self.buff


    def __str__(self):
        return self.buff



class SyncNotebooks(object):
    """
    The SyncNotebooks utility class has a single entrypoint, namely the
    'sync' classmethod used to compute the commands needed to make two
    notebooks consistent.

    The sync method calls commands on the 'dst' notebook needed to make
    it match the 'src' notebook. The commands it uses includes add_cell,
    remove_cell and reorder_cells.

    This utility is used to sync the state of the notebook as declared
    by the editor (i.e emacs) with the state of the notebook in the
    client.
    """


    @classmethod
    def deduplicate_hashes(cls, src, dst):
        """
        Everything mostly works if this simply returns (src.hashes,
        dst.hashes) although cells without prompts and the same source
        will be collapsed into a single cell.

        The solution used here is to deduplicate the hashes in this
        situation as cells without prompts have no output so it does not
        matter that the hashes don't match as there is no output to preserve.
        """
        src_hashes, dst_hashes = src.hashes, dst.hashes
        dedupe_src_hashes, dedupe_dst_hashes = [], []

        src_dupes = set([h for h in src_hashes if src_hashes.count(h) > 1])
        dst_dupes = set([h for h in dst_hashes if dst_hashes.count(h) > 1])

        duplicates = src_dupes | dst_dupes

        for count, (hsh, cell) in enumerate(zip(src_hashes, src.cells)):
            dedupe_src_hashes.append(hsh + count if
                                     (hsh in duplicates and cell.prompt is None) else hsh)

        for count, (hsh, cell) in enumerate(zip(dst_hashes, dst.cells)):
            dedupe_dst_hashes.append(hsh + count if
                                     (hsh in duplicates and cell.prompt is None) else hsh)
        return dedupe_src_hashes, dedupe_dst_hashes


    @classmethod
    def sync(cls, connection, src, dst):
        hashes = cls.deduplicate_hashes(src, dst)
        src_hashes, dst_hashes = hashes

        src_set, dst_set = set(src_hashes), set(dst_hashes)
        additions = src_set - dst_set
        deletions =  dst_set - src_set

        if src_hashes == dst_hashes: # Same cells, same order : no change
            pass
        elif src_set == dst_set: # Same cells, different order
            cls.reorderings(connection, src, dst, hashes)
        elif len(src_set) == len(dst_set):
            cls.updates(connection, src, dst, additions, deletions, hashes)
        elif additions and not deletions:
            cls.additions(connection, src, dst, additions, hashes)
        elif deletions and not additions:
            cls.deletions(connection, dst, deletions, hashes)

        # As this all works via hashes on the stripped source, we need to
        # transfer the exact source lines across to preserve newlines and
        # to correctly infer cell positions from line numbers.
        if len(src.cells) != len(dst.cells):
            logging.info('WARNING: Cell length mismatch %r vs %r' % (src.cells, dst.cells))
        for src_cell, dst_cell in zip(src.cells, dst.cells):
            dst_cell.source = src_cell.source


    @classmethod
    def updates(cls, connection, src, dst, additions, deletions, hashes):
        src_hashes, dst_hashes = hashes

        add_set = {src_hashes.index(h) for h in additions}
        del_set = {dst_hashes.index(h) for h in deletions}
        if add_set.symmetric_difference(del_set):
            logging.info("Sync cannot resolve all cell updates")

        for pos in sorted(add_set.intersection(del_set)):
            dst.update_cell_input(connection,  src.cells[pos].source,
                                  pos, src.cells[pos].input)

    @classmethod
    def reorderings(cls, connection, src, dst, hashes):
        src_hashes, dst_hashes = hashes
        positions = []
        for hsh in dst_hashes:
            position = src_hashes.index(hsh)
            positions.append(position)
        dst.reorder_cells(connection, positions)

    @classmethod
    def additions(cls, connection, src, dst, additions, hashes):
        "When src is the same as dst with some new cells"
        src_hashes, dst_hashes = hashes
        added = []
        for addition in additions:
            ind = src_hashes.index(addition)
            cell = src.cells[ind]
            added.append((ind, cell))

        for ind, cell in sorted(added):
            dst.add_cell(connection, cell.source,
                         cell.input, mode=cell.mode,
                         position=ind, outputs=cell.outputs)

    @classmethod
    def deletions(cls, connection, dst, deletions, hashes):
        "When src is the same as dst with some cells deleted"
        src_hashes, dst_hashes = hashes
        for deletion in deletions:
            ind = dst_hashes.index(deletion)
            dst.remove_cell(connection, ind)



import argparse
from .server import serve

parser = argparse.ArgumentParser()


parser.add_argument('--ws-port', action='store', default=9999, type=int,
                    help='Websocket port for editor and client connections.')

parser.add_argument('--html-port', action='store', default=8000, type=int,
                    help='Port used to serve the HTML content.')

parser.add_argument('--host',
                    action='store',
                    default='127.0.0.1',
                    help='IP address the application should listen on.')

args = parser.parse_args()


def main():
    serve(args.ws_port, args.html_port, args.host)



import os, time, logging, json, uuid, weakref
import atexit

from collections import OrderedDict

import warnings

class VisibleDeprecationWarning(UserWarning):
    """Warning issued by jupyter_client 5.2.4 about future deprecations in tornado"""
    pass

# Hopefully jupyter_client will keep working with latest tornado
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=VisibleDeprecationWarning)
    warnings.simplefilter("ignore")
    from jupyter_client import KernelManager
    from jupyter_client.threaded import ThreadedZMQSocketChannel, ThreadedKernelClient


from nbformat.v4 import output_from_msg
from queue import Queue

class Channel(ThreadedZMQSocketChannel):

    executions = None
    queue = None

    channels = weakref.WeakSet()

    def __init__(self, *args, **kwargs):
        super(Channel, self).__init__(*args, **kwargs)
        self.channels.add(self)

    @classmethod
    def execution_number(cls):
        "Gets the execution count if available on any of the channels"
        counts = [c.executions for c in cls.channels if c.executions]
        return counts[0] if len(counts)==1 else None


    def call_handlers(self, msg):
        msg_type = msg.get('msg_type',None)
        content = msg.get('content', None)
        buffers = msg.get('buffers', [])
        metadata = msg.get('metadata', None)

        if msg_type == 'complete_reply':
            filtered = {}
            for field in ['matches', 'cursor_start', 'cursor_end']:
                if field == 'matches':
                    value = [el for el in content["matches"]
                             if not (el.startswith("%")      # Remove magics
                                     or el.startswith("_"))] # And underscore methods
                else:
                    value = content[field]
                filtered[field] = value

            self.queue.put((filtered, "completion"))
            return

        node = None
        if msg_type == 'execute_input':
            assert 'execution_count' in content
            try:
                executions = int(content['execution_count'])
                self.executions = executions
            except:
                logging.info('ERROR: execution_count is not an integer %r' %
                      content['execution_count'])
                self.executions = None

        elif msg_type == 'error':
            node = output_from_msg(msg)
        # Capture print output
        elif msg_type == 'stream':
            node = output_from_msg(msg)
        elif msg_type == 'status':
            pass # e.g content['execution_state'] == 'idle'

        elif msg_type == "comm_open":
            self.queue.put(({'content':content,
                             'buffers':buffers,
                             'metadata':metadata}, 'comm_open'))
            return
        elif msg_type == "comm_msg":
            self.queue.put(({'content':content,
                             'buffers':buffers,
                             'metadata':metadata}, 'comm_msg'))
            return
        elif msg_type.startswith('comm'):
            logging.info("Unhandled 'comm' message of type {msg_type} with {content}".format(
                msg_type=msg_type,
                content=content))

        if msg_type in {'execute_result', 'display_data', 'update_display_data'}:
            if msg_type == 'update_display_data':
                logging.info("Unhandled 'update_display_data' message")
            # Should filter text/plain and text/html at this level
            # print(output_from_msg(msg)['data'])

            node = output_from_msg(msg)

        if node:
            self.queue.put((node, self.executions))
        elif msg_type == 'execute_reply':
            self.queue.put((None, self.execution_number()))


class Client(ThreadedKernelClient):
    iopub_channel_class = Channel
    shell_channel_class = Channel
    stdin_channel_class = Channel


class ThreadedExecutor(object):

    def __init__(self, name, queue=Queue()):
        self.name = name

        self.comms = Comms(self)
        self.km = None
        self.kc = None

        self.execution_count = 0
        self.executable = None

        Channel.queue = queue

    def start(self, cwd=None, executable=None):
        self.km = KernelManager(kernel_name='python',
                                client_class='nei.execute.Client')
        self.executable = executable
        if executable is not None:
            self.km.kernel_spec.argv[0] = executable
        self.km.start_kernel(**({} if cwd is None else {'cwd':cwd}))
        self.kc = self.km.client()
        self.kc.start_channels()
        time.sleep(2)
        atexit.register(self.shutdown_kernel)

    def reset(self):
        Channel.executions = 0

    def restart_kernel(self):
        self.execution_count = 0
        self.km.restart_kernel()
        Channel.executions = 0

    def interrupt_kernel(self):
        self.km.interrupt_kernel()

    def shutdown_kernel(self): # TODO: Shutdown kernel but keep nei running
        self.execution_count = 0
        self.km.request_shutdown()
        self.km.cleanup()
        self.km.finish_shutdown()
        logging.info('Shutting down')

    def kernel_info(self):
        self.kc.kernel_info()

    def complete(self, code, position):
        self.kc.complete(code, position)

    def __call__(self, code, stop_on_error=True, cell=None, silent=False):
        "If stop_on_error is True, execution may stop on exceptions"

        if not silent:
            self.execution_count += 1

        if cell:
            cell.prompt = self.execution_count
        self.kc.execute(code,
                        silent=silent,
                        store_history=True, # Has to be true to make execution counter work
                        allow_stdin=False,
                        stop_on_error=stop_on_error)


"""
NOTES

A RemoteExecute kernel could connect via a port to a remote kernel instance:

http://www.giantflyingsaucer.com/blog/?p=4602
http://www.tornadoweb.org/en/stable/websocket.html#tornado.websocket.websocket_connect
https://github.com/ilkerkesen/tornado-websocket-client-example/blob/master/client.py

Is get_iopub_msg a useful method?
"""


class Comms(object):

    def __init__(self, executor):
        self.executor = executor
        self.opened = set()

    def _publish_msg(self, cmd, comm_id, data=None, metadata=None, buffers=None, **keys):
        target_module = '' # requirejs module from which to load comm target
        content = dict(data        = data if data else {},
                       comm_id     = comm_id,
                       buffers     = buffers, # TODO: Needed for Bokeh
                       **keys)  # Includes target_name

        session = self.executor.km.session
        shell_channel_socket = self.executor.kc.shell_channel.socket
        session.send(shell_channel_socket, cmd, json.dumps(content))

    def comm_open(self, comm_id, target_name, data=None, metadata=None,
                  buffers=None, **keys):
        self.opened.add(comm_id)
        self._publish_msg('comm_open', comm_id, target_name=target_name,
                          data=data, metadata=metadata, buffers=buffers)

    def comm_msg(self, comm_id, target_name, data=None, metadata=None
                 , buffers=None, **keys):
        logging.info("Sending message to Python comm %s." % comm_id)
        self._publish_msg('comm_msg', comm_id, target_name=target_name,
                          data=data, metadata=metadata, buffers=buffers)



import pkgutil
try:
    import cssutils
except:
    cssutils = None

def parse_CSS(css):
    """
    Given some CSS, parse it and return a dictionary of dictionaries:

    { selector : {property: value}}
    """
    sheet = cssutils.parseString(css)
    css_dict={}
    for rule in sheet:
        selector = rule.selectorText
        styles = rule.style.cssText
        prop_dict = {}
        for prop in rule.style:
            prop_dict[prop.name] = prop.value
        css_dict[selector] = prop_dict
    return css_dict



def css_processor(code_generator):
    """
    Create a CSS processor based on the given code generator.
    """
    def inner(notebook, connection, css):
        if cssutils is None:
            print('WARNING: cssutil is not available to handle CSS')
            return
        code = code_generator(parse_CSS(css))
        if code is not None:
            notebook.exec_silently(connection, code)
    return inner



def apply_holoviews_theme(parsed_css):
    """
    A code generator used to set the holoviews theme from the parsed CSS.
    """
    hvavailable = pkgutil.find_loader('holoviews')
    if hvavailable is None: return
    if parsed_css == {}:
        return ''

    lines = ["import holoviews as hv",
             "from holoviews.plotting import bokeh",
             "from bokeh.themes.theme import Theme",
             "hv.renderer('bokeh').theme = Theme(json={json})"]

    bgcolor = parsed_css['body']['background-color']
    bgcolor = bgcolor if bgcolor != '#fff' else  '#ffffff'
    opposed = 'white' if bgcolor != '#ffffff' else 'black'
    theme = {
            'attrs' : {
                'Figure' : {
                    'background_fill_color': bgcolor,
                    'border_fill_color':     bgcolor,
                    'outline_line_color': '#444444',
                },
                'Grid': {
                    'grid_line_dash': [6, 4],
                    'grid_line_alpha': .3,
                },
                'Title': {
                    'text_color': opposed
                },
                'Axis': {
                    'axis_label_text_color': opposed,
                    'axis_line_color' : opposed,
                    'minor_tick_line_color' : opposed,
                    'major_label_text_color' : opposed,
                    'major_tick_line_color': opposed
                },
                'ColorBar': {
                    'background_fill_color' : bgcolor,
                    'title_text_color': opposed,
                    'major_label_text_color': opposed
                    }
            }
        }

    template = "\n".join(lines)
    return template.format(json=repr(theme))



process_css = css_processor(apply_holoviews_theme)



import os, sys
from .server import serve, WS, session

__version__ = 'v0.0.7'


def server_status(port, host='localhost'):
    import socket
    from contextlib import closing
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        if sock.connect_ex((host, port)) == 0:
            print("port unavailable")



import re
import logging



class ValidatingParser(object):

    @classmethod
    def truncate(cls, source):
        if len(source) > 40:
            return source[:20] + '...' + source[-20:]
        else:
            return source

    @classmethod
    def validate_notebook(cls, nb, dict_cells, buffer_text, cell_cls):
        """
        Used to validate parsing process with a notebook ipynb file is
        loaded by comparing the notebook object to the parse result. If
        validation fails, returns an error message otherwise returns
        None.
        """
        ipynb_modes = [c['cell_type'] for c in nb.cells]
        ipynb_sources =  [c['source'] for c in nb.cells]
        if  ipynb_modes != [c['mode'] for c in dict_cells]:
            print("Very surprising! Not redundant?")
        if ipynb_sources != [c['source'] for c in dict_cells]:
            print("Even more surprising! Not redundant?")

        parsed_cells = cls.extract_cells(buffer_text, cell_cls, unique_prompts=True)
        parsed_modes = [c['mode'] for c in parsed_cells]
        parsed_sources = [c['source'] for c in parsed_cells]
        if ipynb_modes != parsed_modes:
            logging.info('!!WARNING!!\nMismatch between JSON and parsed cell types:\n'
                         + ('\tIPYNB: %s\n' % ipynb_modes)
                         + ('\tPARSED: %s' % parsed_modes))
            return False


        for ipy_source, parse_source in zip(ipynb_sources, parsed_sources):
            if ipy_source!=parse_source:
                lenmsg = (('!!WARNING!!\nMismatch between JSON source [length %d] and '
                          'parsed cell source [length %d]:\n')
                          % (len(ipy_source), len(parse_source)))
                logging.info(lenmsg
                         + ('\tIPYNB: %r\n' % cls.truncate(ipy_source)
                         + ('\tPARSED: %r' %  cls.truncate(parse_source))))
                return False
        return True


class ParseNotebook(ValidatingParser):

    @classmethod
    def find_markdown_range(cls, start, stop, lines):
        "Look backwards from stop index to start index for markdown opener"
        md_start = None
        for i in list(range(stop-1, start-1,-1)):
            if lines[i].startswith("\"\"\""): # REGEXP TO TERMINATE
                md_start = i
                break

        return None if md_start is None else (md_start, stop, 'markdown', {})

    @classmethod
    def filter_out_markdown_ranges(cls, lines, ranges):
        """
        As markdown is the first pass so all lines should be present.

        This function filters out the identified markdown ranges from all
        the lines to find out what is remaining (potential code cells).
        """
        all_line_numbers = list(range(len(lines)))
        accounted_lines = [list(range(start, stop+1)) for (start, stop, _, _) in ranges]
        flattened = [el for group in accounted_lines for el in group]
        remaining_line_numbers = set(all_line_numbers) - set(flattened)
        return [(el, lines[el]) for el in sorted(remaining_line_numbers)]

    @classmethod
    def find_markdown_ranges(cls, lines):
        """
        Return a list of range tuples (start, stop, 'md') and a list of
        remaining (numbered) lines
        """
        numbered_lines = list(enumerate(lines))
        md_ranges = []
        start = 0
        for ind, line in numbered_lines:
            if line.startswith("\"\"\" #:md:"): # REGEXP TO TERMINATE
                md_range = cls.find_markdown_range(start, ind, lines)
                if md_range is None: continue
                else:
                    (start, _, _, _ ) = md_range
                    md_ranges.append(md_range)

        return md_ranges, cls.filter_out_markdown_ranges(lines, md_ranges)

    @classmethod
    def find_code_ranges(cls, lines, code_prompt_regexp):
        """
        Find the ranges for the code cells.

        Works by first identifying the opening lines for code cells. Then it
        iterates over the ranges between these boundaries, breaking as soon
        as a line is not available (i.e an intermediate markdown cell).
        """
        code_starts = [(re.match(code_prompt_regexp, line), line_no)
                       for (line_no, line) in lines
                       if re.match(code_prompt_regexp, line)]
        if len(code_starts) == 0:
            return []
         # Add maximum available line number
        code_starts = code_starts + [(None, lines[-1][0])]
        available_lines = set([line_no for (line_no,_) in lines])
        code_ranges = []
        for start_info, end_info in zip(code_starts[:-1], code_starts[1:]):
            (_, end) = end_info
            (match, start) = start_info
            raw_prompt = match.groups()[0]
            try:
                prompt = int(raw_prompt) if raw_prompt not in [None, ' '] else None
            except Exception as e:
                logging.info('WARNING: Parse failure for prompt count %r' % str(e))
                prompt = None

            code_range = (start, start)
            for test_index in range(start, end+1):
                if  (test_index in available_lines):
                    code_range = (start, test_index, 'code', {'prompt':prompt})
                else:
                    code_ranges.append(code_range)
                    break
            else:
                code_ranges.append(code_range)

        return cls._adjust_code_range_boundaries(code_ranges)

    @classmethod
    def _adjust_code_range_boundaries(cls, code_ranges):
        """
        Fix boundaries between successive code cells as lines ranges
        include next prompt line otherwise.
        """
        adjusted_ranges = []
        for i, crange in enumerate(code_ranges):
            if i < (len(code_ranges)-1):
                (start, end, _, info) = crange
                next_crange = code_ranges[i+1]
                (next_start, next_end, _, _) = next_crange
                if end == next_start:
                    crange = (start, end-1, 'code', info)
            adjusted_ranges.append(crange)
        return adjusted_ranges

    @classmethod
    def collect_cells(cls, lines, ranges):
        """
        Given the set of lines and the identified cell ranges, extract the
        cells.
        """
        cells = []
        sorted_ranges = sorted(ranges)
        for (start, stop, cell_type, info) in sorted_ranges:
            source = "".join(lines[start+1:stop+1])
            md_terminator = "\n\"\"\" #:md:"
            # FIXME! Use regexp search replace when any number of newlines at end of source.
            if cell_type == "markdown" and source.endswith(md_terminator):
                source = source[:-len(md_terminator)]
            elif cell_type == "markdown" and source.endswith(md_terminator+'\n'):
                source = source[:-(len(md_terminator)+1)]
            elif ((stop != (len(lines)-1))
                  and cell_type == "code" and source.endswith("\n\n")):
                source = source[:-2]

            cells.append(dict({'mode':cell_type,
                               'source':source}, **info))
        return cells

    @classmethod
    def unique_prompts(cls, cells):
        "Ensures prompts are either unique or None"
        unique, seen = [], []
        for cell in cells:
            if cell['mode'] != 'code':
                unique.append(cell)
                continue
            elif cell['prompt'] in seen:
                cell['prompt'] = None
            else:
                seen.append(cell['prompt'])
            unique.append(cell)
        return unique

    @classmethod
    def extract_cells(cls, source, cell_cls, unique_prompts=True): # FIX cell_cls ARG
        lines =source.splitlines(keepends=True)
        md_ranges, remaining_lines = cls.find_markdown_ranges(lines)
        code_ranges = cls.find_code_ranges(remaining_lines, cell_cls.code_prompt_regexp)
        extracted = cls.collect_cells(lines, md_ranges+code_ranges)
        return cls.unique_prompts(extracted) if unique_prompts else extracted


if __name__ == '__main__':
    print(ParseNotebook.extract_cells(source, None))



#!/usr/bin/env python

import logging
import tornado
import tornado.web
from tornado import httpserver
from tornado import ioloop
from tornado import websocket

import os
import sys
import json
import webbrowser
import nbformat
from queue import Queue

from .execute import ThreadedExecutor
from .cells import ExecutableNotebook

STATIC_PATH = os.path.join(os.path.split(__file__)[0], 'client')
ExecutableNotebook.STATIC_PATH = STATIC_PATH

class Session(object):
    """
    Global state of the server that doesn't belong to any single
    websocket connection but to the whole server session.
    """
    def __init__(self):
        self._browser = None
        self._editor = None
        self.notebook = None
        self.buffers = {}

    def reset(self):
        self._browser = None
        self.editors = []
        self.notebook = None
        self.buffers = {}

    @property
    def browser(self):
        return self._browser

    @browser.setter
    def browser(self, connection):
        if self._browser is not None:
            logging.info("WARNING: Only one browser connection expected")
        self._browser = connection

    @property
    def editor(self):
        return self._editor

    @editor.setter
    def editor(self, connection):
        if self._editor is not None:
            logging.info("WARNING: Only editor browser connection expected")
        self._editor = connection


session = Session()

class PeriodicOutputCallback(object):
    """
    Sets up a periodic callback to push output to cells by polling from
    the queue pushed to by the ThreadedExecutor.
    """

    def __init__(self, server, period=20):
        self.server = server
        self.notebook = None
        self.period = period

    def switch_notebook(self, notebook):
        self.notebook = notebook

    def start(self):
        self.callback = ioloop.PeriodicCallback(self.__call__, self.period)
        self.callback.start()

    def stop(self):
        self.callback.stop()

    def __call__(self):
        "Processes queue pushed to by ThreadedExecutor"
        try:
            val = self.server.queue.get_nowait()
            self.server.queue.task_done()
            result, status  = val
        except:
            return

        if status == 'completion':
            position = self.notebook.completion_info['position']
            relative_position = self.notebook.completion_info['relative_position']

            # Adjusted for emacs point position
            start_delta = relative_position - result['cursor_start']
            end_delta = relative_position - result['cursor_end']
            result['cursor_start'] = position - start_delta
            result['cursor_end'] = position - end_delta

            session.editor.write_message(json.dumps({'cmd':'completion',
                                                     'data': result}))
            self.notebook.completion_info = None
            return

        if session.browser and (status == 'comm_open'):
            logging.info("REQUEST TO OPEN COMM FOR JS: %s" % result)
            self.notebook.message(session.browser, 'comm_open', result['content'])
            # e.g:
            # {'data': {}, 'comm_id': 'ee0a39d3728945cdb4ad30848b7856fc',
            #  'target_name': 'ZOO', 'target_module': None}
            return
        elif session.browser and (status == 'comm_msg'):
            buffers = result['buffers']
            metadata = result.get('metadata', {})
            self.notebook.message(session.browser, 'comm_msg', # FIXME: redundant 'comm_msg'
                                  {'msg_type': 'comm_msg',
                                   'metadata': metadata,
                                   'content': result['content']},
                                  buffers=buffers)
            return


        else:
            outnode, execution_count = result, status

        if session.browser:
            cell = self.notebook.find_cell(execution_count)
            if cell is None: return # There may be no cell if running a silent execution
            position = self.notebook.cell_position(cell)

            if execution_count is None:
                # For silent execution before *any* output
                return
            # What about silent execution after start?
            self.notebook.update_cell_outputs(session.browser, position, outnode)


class WS(websocket.WebSocketHandler):

    def open(self):
        self.queue = Queue()
        self.output_callback = PeriodicOutputCallback(self)
        self.output_callback.start()
        logging.info("Connection opened")


    def toggle_notebook(self, name):
        notebook = session.buffers.get(name, None)

        if notebook is None:  # Create notebook
            # Note that there are multiple connections and we want only one notebook!
            # (for now)
            notebook = ExecutableNotebook(
                (ThreadedExecutor, "threaded-kernel", self.queue),
                name=name, cells=list())
            session.buffers[name] = notebook

        session.notebook = notebook
        self.output_callback.switch_notebook(notebook)


    def on_message(self, message):
        "Websocket on_message handler. Tracks connection type."
        try:
            payload = json.loads(message)
        except Exception as e:
            logging.info('JSON parse exception: %s' % str(e))
            return

        if 'cmd' in payload:
            if payload['cmd'] in ['start_mirror']: # Verbose commands
                logging.info(u"Received %s command" % payload['cmd'])
            else:
                logging.info(u"Received message: {0:<.100}".format(message))

        if payload.get('cmd') == 'reset_server':
            self.output_callback.stop()
            session.reset()
            return

        if payload.get('init', False) == 'editor':
            logging.info('Added editor client connection')
            session.editor = self
            return

        if payload.get('init', False) == 'browser':
            session.browser = self
            logging.info('Added browser client connection')
            if session.notebook and len(session.notebook.cells) > 0: # TODO: Needs updating
                logging.info("Restart with previously opened notebook")
                session.notebook.reload(self)
                # If you hit reload in the browser, the CSS needs to be re-sent
                session.notebook.update_theme(self, css=None)
            return

        # SOME COMMANDS (e.g mirroring) should happen even without a browser tab open!
        self.toggle_notebook(payload['name'])

        if payload.get('cmd', False) == 'reload_page':
            # Reload over the browser connection (currently assuming only one)
            if session.browser is not None:
                session.notebook.reload(session.browser)
            return

        editor_msg = session.notebook.dispatch(session.browser, payload)
        if (editor_msg is not None) and (session.editor is not None):
            session.editor.write_message(json.dumps(editor_msg))


    def check_origin(self, origin):
        return True


    def on_close(self):
        logging.info("ON_CLOSE")
        if self is session.browser:
            session.browser = None

        self.output_callback.stop()

def serve(ws_port=9999, html_port=8000, host='127.0.0.1'):
    import logging
    logging.basicConfig(level=logging.INFO)
    html_handler = (r'/(.*)', tornado.web.StaticFileHandler,
                    {'path': STATIC_PATH})
    tornado.web.Application([html_handler]).listen(html_port)
    ws_server = httpserver.HTTPServer(tornado.web.Application([(r"/", WS)]))
    ws_server.listen(ws_port, host)
    logging.info("STARTED: Server started and listening")
    ioloop.IOLoop.instance().start()



from setuptools import setup

install_requires = [
    'jupyter_client==5.2.4',
    'nbformat',
    'nbconvert',
    'tornado',
    'ipykernel'
]

extras_require = {
    'pyviz': [
        'bokeh',
        'holoviews',
        'cssutils'
    ]
}

setup_args = dict(
    name='nei',
    version='0.0.7',
    description='Notebook Emacs Interface',
    platforms=['Mac OS X', 'Linux'],
    license='BSD',
    provides=["nei"],
    packages=["nei"],
    include_package_data = True,
    python_requires=">=3.5",
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        'console_scripts': [
            'nei = nei.cmd:main'
        ]},
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English"
    ]
)

if __name__=="__main__":
    setup(**setup_args)
